import geopandas as gpd
import networkx as nx
import pandas as pd
from shapely import LineString, Point
from shapely.ops import substring

from iduedu import config
from iduedu.utils.utils import remove_weakly_connected_nodes

logger = config.logger


def join_pt_walk_graph(public_transport_g: nx.Graph, walk_g: nx.Graph, max_dist=20) -> nx.Graph:
    """
    Combine a public transport network graph with a pedestrian network graph, creating an intermodal transport graph.
    Platforms in the public transport network are connected to nearby pedestrian network edges based on the specified
    maximum distance.

    Parameters
    ----------
    public_transport_g : nx.Graph
        The public transport network graph, which should contain platform nodes with geometry and CRS information.
    walk_g : nx.Graph
        The pedestrian network graph. It must have the same CRS as the public transport graph.
    max_dist : float, optional
        Maximum distance (in meters) to search for connections between platforms and pedestrian edges. Defaults to 20.

    Returns
    -------
    nx.Graph
        A combined intermodal graph where public transport platforms are connected to nearby pedestrian routes.

    Raises
    ------
    AssertionError
        If the CRS of the public transport graph and pedestrian graph do not match.

    Examples
    --------
    >>> intermodal_graph = join_pt_walk_graph(public_transport_g, walk_g, max_dist=50)

    Notes
    -----
    The function relabels nodes in both graphs to ensure unique node IDs before composing them.
    It connects public transport platforms to the closest edges in the pedestrian network by projecting platforms
    onto edges.
    Walking speed is taken from the pedestrian graph, and default speed is set to 83.33 m/min if not available.
    """

    assert public_transport_g.graph["crs"] == walk_g.graph["crs"], "CRS mismatching."
    logger.info("Composing intermodal graph...")

    num_nodes_g1 = len(public_transport_g.nodes)
    mapping_g1 = {node: idx for idx, node in enumerate(public_transport_g.nodes)}
    mapping_g2 = {node: idx + num_nodes_g1 for idx, node in enumerate(walk_g.nodes)}
    transport: nx.DiGraph = nx.relabel_nodes(public_transport_g, mapping_g1)
    walk: nx.MultiDiGraph = nx.relabel_nodes(walk_g, mapping_g2)

    platforms = pd.DataFrame.from_dict(dict(transport.nodes(data=True)), orient="index")
    platforms = platforms[platforms["type"] == "platform"]
    platforms["geometry"] = gpd.points_from_xy(platforms["x"], platforms["y"])
    platforms = gpd.GeoDataFrame(platforms, crs=transport.graph["crs"])

    logger.debug("creating walk edges gdf")
    walk_edges = gpd.GeoDataFrame(
        nx.to_pandas_edgelist(walk.to_undirected(), source="u", target="v", edge_key="k"), crs=walk.graph["crs"]
    )
    walk_edges["edge_geometry"] = walk_edges["geometry"]

    logger.debug("joining nearest")
    projection_join = platforms.reset_index().sjoin_nearest(walk_edges, max_distance=max_dist, distance_col="dist")
    projection_join["project_dist"] = projection_join["edge_geometry"].project(projection_join["geometry"])
    projection_join["project_point"] = projection_join["edge_geometry"].interpolate(projection_join["project_dist"])
    projection_join["route"] = projection_join["route"].apply(lambda x: x if isinstance(x, list) else [x])

    logger.debug("searching for duplicated project points")
    projection_join = projection_join.groupby(by="project_point", as_index=False).agg(
        {
            "index": tuple,
            "index_right": "first",
            "geometry": "first",
            "u": "first",
            "v": "first",
            "k": "first",
            "project_dist": "first",
            "route": lambda x: [item for sublist in x for item in sublist],
        }
    )
    for ind, row in projection_join.iterrows():
        if len(row["index"]) == 1:
            projection_join.loc[ind, "index"] = row["index"][0]
        platform_in = row["index"][0]
        for platform_out in row["index"][1:]:
            if platform_out == platform_in:
                continue
            for neighbor in transport.neighbors(platform_out):
                edge_data = transport.get_edge_data(neighbor, platform_out)
                if edge_data:
                    transport.add_edge(neighbor, platform_in, **edge_data)
                edge_data = transport.get_edge_data(platform_out, neighbor)
                if edge_data:
                    transport.add_edge(platform_in, neighbor, **edge_data)
            transport.remove_node(platform_out)
        transport.nodes[platform_in]["route"] = row["route"]
        projection_join.loc[ind, "index"] = platform_in

    logger.debug("Projecting platforms to walk graph")
    points_grouped_by_edge = projection_join.groupby(by="index_right", as_index=False).agg(
        {
            "index": tuple,
            "geometry": tuple,
            "u": "first",
            "v": "first",
            "k": "first",
        }
    )
    try:
        speed = walk.graph["walk_speed"]
    except KeyError:
        logger.warning(
            "There is no walk_speed in graph, set to the default speed - 83.33 m/min"
        )  #  посчитать примерную скорость по length timemin для любой эджи
        speed = 83.33

    edges_to_del = []

    for _, row in points_grouped_by_edge.iterrows():
        u, v, k = row[["u", "v", "k"]]
        edge: LineString = walk.get_edge_data(u, v, k)["geometry"]
        # Если платформа одна единственная на эдж
        if len(row["index"]) == 1:
            platform_id = row["index"][0]
            dist = edge.project(row.geometry[0])
            projected_point = edge.interpolate(dist)
            if dist == 0:
                # Если платформа проецируется на начало эджа
                mapping = {u: platform_id}
                nx.relabel_nodes(walk, mapping, copy=False)
            elif dist == edge.length:
                # Если на конец
                mapping = {v: platform_id}
                nx.relabel_nodes(walk, mapping, copy=False)
            else:
                line1 = substring(edge, 0, dist)
                line2 = substring(edge, dist, edge.length)
                # Убираем старые эджи и добавляем новые, в серединке нода - наша платформа для соединения
                edges_to_del.append((u, v, k))
                if u != v:
                    edges_to_del.append((v, u, k))
                walk.add_node(platform_id, x=projected_point.x, y=projected_point.y)
                walk.add_edge(
                    u,
                    platform_id,
                    geometry=line1,
                    length_meter=round(line1.length, 3),
                    time_min=round(line1.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    platform_id,
                    u,
                    geometry=line1.reverse(),
                    length_meter=round(line1.length, 3),
                    time_min=round(line1.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    platform_id,
                    v,
                    geometry=line2,
                    length_meter=round(line2.length, 3),
                    time_min=round(line2.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    v,
                    platform_id,
                    geometry=line2.reverse(),
                    length_meter=round(line2.length, 3),
                    time_min=round(line2.length / speed, 3),
                    type="walk",
                )
        # Если платформм несколько на эдж
        else:
            dist_project = [
                (edge.project(platform), edge.interpolate(edge.project(platform)), ind)
                for ind, platform in zip(row["index"], row.geometry)
            ]
            dist_project.sort(key=lambda x: x[0])
            u_to_del = u
            v_to_del = v
            last_dist = 0
            last_u = u
            for dist, projected_point, cur_index in dist_project:
                if dist == 0:
                    # Если платформа проецируется на начало эджа
                    mapping = {u: cur_index}
                    u_to_del = cur_index
                    nx.relabel_nodes(walk, mapping, copy=False)
                elif dist == edge.length:
                    # Если на конец
                    mapping = {v: cur_index}
                    v_to_del = cur_index
                    nx.relabel_nodes(walk, mapping, copy=False)
                else:
                    # По очереди добавляем ноды/линии
                    line = substring(edge, last_dist, dist)
                    if isinstance(line, Point):
                        raise ValueError(f"wtf {line} cannot be linestring")
                    walk.add_node(cur_index, x=projected_point.x, y=projected_point.y)
                    walk.add_edge(
                        last_u,
                        cur_index,
                        geometry=line,
                        length_meter=round(line.length, 3),
                        time_min=round(line.length / speed, 3),
                        type="walk",
                    )
                    walk.add_edge(
                        cur_index,
                        last_u,
                        geometry=line.reverse(),
                        length_meter=round(line.length, 3),
                        time_min=round(line.length / speed, 3),
                        type="walk",
                    )
                    last_u = cur_index
                last_dist = dist

            # Если последняя остановка спроецировалась не на конец эджа, надо добавить остаток
            if last_dist != edge.length:
                line = substring(edge, last_dist, edge.length)
                walk.add_edge(
                    last_u,
                    v,
                    geometry=line,
                    length_meter=round(line.length, 3),
                    time_min=round(line.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    v,
                    last_u,
                    geometry=line.reverse(),
                    length_meter=round(line.length, 3),
                    time_min=round(line.length / speed, 3),
                    type="walk",
                )
            edges_to_del.append((u_to_del, v_to_del, k))
            if u != v:
                edges_to_del.append((v_to_del, u_to_del, k))
    walk.remove_edges_from(edges_to_del)
    logger.debug("Composing graphs")
    intermodal = nx.compose(nx.MultiDiGraph(transport), nx.MultiDiGraph(walk))
    intermodal = remove_weakly_connected_nodes(intermodal)
    intermodal.remove_nodes_from([node for node, data in intermodal.nodes(data=True) if "x" not in data.keys()])
    mapping = {old_label: new_label for new_label, old_label in enumerate(intermodal.nodes())}
    logger.debug("Relabeling")
    nx.relabel_nodes(intermodal, mapping, copy=False)
    intermodal.graph["type"] = "intermodal"
    logger.debug("Done composing!")
    return intermodal



import math
import warnings

import numpy as np
import pandas as pd
from pandas import DataFrame
from pyproj import CRS, Transformer
from scipy.spatial import cKDTree
from scipy.spatial.distance import cdist
from shapely import Point
from shapely.geometry import LineString
from shapely.ops import substring

PLATFORM_ROLES = ["platform_entry_only", "platform", "platform_exit_only"]
STOPS_ROLES = ["stop", "stop_exit_only", "stop_entry_only"]


def _link_unconnected(disconnected_ways) -> list:
    # Считаем связи между линиями
    connect_points = [point for coords in disconnected_ways for point in (coords[0], coords[-1])]
    distances = cdist(connect_points, connect_points)
    n = distances.shape[0]
    mask = (np.arange(n)[:, None] // 2) == (np.arange(n) // 2)
    distances[mask] = np.inf

    def relative_point(point: int):
        rel_point = point // 2 * 2
        return rel_point if rel_point != point else rel_point + 1

    connected_ways = []
    # pylint: disable=unbalanced-tuple-unpacking
    first_con_1, first_con_2 = np.unravel_index(np.argmin(distances), distances.shape)
    distances[first_con_1, :] = np.inf
    distances[first_con_2, :] = np.inf
    distances[:, first_con_2] = np.inf
    distances[:, first_con_1] = np.inf
    first_con_1_rel, first_con_2_rel = [relative_point(x) for x in (first_con_1, first_con_2)]
    distances[first_con_1_rel, first_con_2_rel] = np.inf
    distances[first_con_2_rel, first_con_1_rel] = np.inf
    distances[:, first_con_1_rel] = np.inf
    distances[:, first_con_2_rel] = np.inf

    line1, line2 = first_con_1 // 2, first_con_2 // 2

    if first_con_1 % 2 == 1:
        connected_ways += disconnected_ways[line1]
    else:
        connected_ways += disconnected_ways[line1][::-1]
    if first_con_2 % 2 == 0:
        connected_ways += disconnected_ways[line2]
    else:
        connected_ways += disconnected_ways[line2][::-1]

    extreme_points = [first_con_1_rel, first_con_2_rel]

    for _ in range(len(disconnected_ways) - 2):
        # pylint: disable=invalid-sequence-index
        position, ind = np.unravel_index(np.argmin(distances[extreme_points]), (2, n))
        next_con = (extreme_points[position], ind)
        rel_point = relative_point(next_con[1])

        line = ind // 2
        if position == 0:
            connected_ways = (
                disconnected_ways[line] + connected_ways
                if ind % 2 == 1
                else (disconnected_ways[line][::-1] + connected_ways)
            )
            extreme_points = [rel_point, extreme_points[1]]
        else:
            connected_ways = (
                connected_ways + disconnected_ways[line]
                if ind % 2 == 0
                else (connected_ways + disconnected_ways[line][::-1])
            )
            extreme_points = [extreme_points[0], rel_point]

        # Чтоб нельзя было зациклиться
        distances[:, rel_point] = np.inf
        distances[next_con[0], :] = np.inf
        distances[next_con[1], :] = np.inf
        distances[:, next_con[0]] = np.inf
        distances[:, next_con[1]] = np.inf
        rel_point_2 = relative_point(next_con[0])
        distances[rel_point, rel_point_2] = np.inf
        distances[rel_point_2, rel_point] = np.inf

    return connected_ways


def parse_overpass_route_response(loc: dict, crs: CRS) -> pd.Series:
    transformer = Transformer.from_crs("EPSG:4326", crs, always_xy=True)

    def transform_geometry(loc):
        if isinstance(loc["geometry"], float):
            return transformer.transform(loc["lon"], loc["lat"])
        p = LineString([transformer.transform(coords["lon"], coords["lat"]) for coords in loc["geometry"]]).centroid
        return p.x, p.y

    def process_roles(route, roles):
        filtered = route[route["role"].isin(roles)]
        if len(filtered) == 0:
            return None
        return filtered.apply(transform_geometry, axis=1).tolist()

    if "ref" in loc["tags"].keys():
        transport_name = loc["tags"]["ref"]
    elif "name" in loc["tags"].keys():
        transport_name = loc["tags"]["name"]
    else:
        transport_name = None
    route = pd.DataFrame(loc["members"])

    if "geometry" not in route.columns:
        route["geometry"] = np.nan
    if "lat" not in route.columns:
        route["lat"] = np.nan
    if "lon" not in route.columns:
        route["lon"] = np.nan

    route = route.dropna(subset=["lat", "lon", "geometry"], how="all")

    platforms = process_roles(route, PLATFORM_ROLES)
    stops = process_roles(route, STOPS_ROLES)

    ways = route[(route["type"] == "way") & (route["role"] == "")]

    if len(ways) > 0:
        ways = ways["geometry"].reset_index(drop=True)
        ways = ways.apply(lambda x: ([transformer.transform(coords["lon"], coords["lat"]) for coords in x])).tolist()
        connected_ways = [[]]
        cur_way = 0
        for coords in ways:
            # Соединяем маршруты, если всё ок, идут без пропусков
            if not connected_ways[cur_way]:
                connected_ways[cur_way] += coords
                continue

            if coords[0] == coords[-1]:
                # Круговое движение зацикленное зачастую в осм, можно отработать, но сходу не придумал
                continue
            if connected_ways[cur_way][-1] == coords[0]:
                connected_ways[cur_way] += coords[1:]
            elif connected_ways[cur_way][-1] == coords[-1]:
                connected_ways[cur_way] += coords[::-1][1:]
            elif connected_ways[cur_way][0] == coords[0]:
                connected_ways[cur_way] = coords[1:][::-1] + connected_ways[cur_way]
            elif connected_ways[cur_way][0] == coords[-1]:
                connected_ways[cur_way] = coords + connected_ways[cur_way][1:]
            # Случай если нету соединяющей точки между линиями маршрута
            else:
                connected_ways += [coords]
                cur_way += 1
        # Соединяем линии по ближайшим точкам этих линий
        if len(connected_ways) > 1:
            to_del = [i for i, data in enumerate(connected_ways) if data[0] == data[-1]]
            connected_ways = [i for j, i in enumerate(connected_ways) if j not in to_del]
        if len(connected_ways) > 1:
            connected_ways = _link_unconnected(connected_ways)
        else:
            connected_ways = connected_ways[0]

    else:
        connected_ways = None

    return pd.Series({"path": connected_ways, "platforms": platforms, "stops": stops, "route": transport_name})


def geometry_to_graph_edge_node_df(loc: pd.Series, transport_type, loc_id) -> DataFrame | None:
    graph_data = []
    node_id = 0
    name = loc.route
    last_dist = None
    last_projected_stop_id = None
    platforms = loc.platforms
    stops = loc.stops
    path = loc.path

    def add_node(desc, x, y, transport=None):
        x = round(x, 5)
        y = round(y, 5)
        if not transport:
            graph_data.append({"node_id": (loc_id, node_id), "point": (x, y), "route": name, "type": desc})
        else:
            graph_data.append({"node_id": (loc_id, node_id), "point": (x, y), "route": name, "type": transport_type})

    def add_edge(u, v, geometry=None, transport=None):
        if not transport:
            graph_data.append(
                {
                    "u": (loc_id, u),
                    "v": (loc_id, v),
                    "route": name,
                    "type": "boarding",
                }
            )
        else:
            graph_data.append(
                {
                    "u": (loc_id, u),
                    "v": (loc_id, v),
                    "geometry": LineString([(round(x, 5), round(y, 5)) for x, y in geometry.coords]),
                    "route": name,
                    "type": transport_type,
                }
            )

    def offset_direction(point):
        # 1 if left 0 if right для определения с какой стороны от линии точки
        dist = path.project(point)

        d1 = dist - 1 if dist - 1 > 0 else 0
        d2 = dist + 1 if dist + 1 < path.length else path.length
        line = substring(path, d1, d2)
        x1, y1 = line.coords[0]
        x2, y2 = line.coords[-1]
        x, y = point.coords[0]

        cross_product = (x2 - x1) * (y - y1) - (y2 - y1) * (x - x1)
        if cross_product > 0:
            return 1
        return 0

    def offset_point(point, direction, distance=5):
        # для размещения платформы по одну сторону от пути на расстоянии
        dist = path.project(point)
        d1 = dist - 1 if dist - 1 > 0 else 0
        d2 = dist + 1.1 if dist + 1.1 < path.length else path.length
        nearest_pt_on_line = path.interpolate(dist)
        line = substring(path, d1, d2)

        x1, y1 = line.coords[0]
        x2, y2 = line.coords[-1]

        dx, dy = x2 - x1, y2 - y1
        length = math.sqrt(dx**2 + dy**2)
        dx, dy = dx / length, dy / length

        if direction == 0:  # Вправо
            nx, ny = dy, -dx
        else:  # Влево
            nx, ny = -dy, dx

        # Смещенная точка
        offset_x = nearest_pt_on_line.x + nx * distance
        offset_y = nearest_pt_on_line.y + ny * distance
        return offset_x, offset_y

    def process_platform(platform):
        nonlocal node_id, last_dist, last_projected_stop_id

        dist = path.project(platform)
        projected_stop = path.interpolate(dist)
        add_node("stop", projected_stop.x, projected_stop.y, transport=True)
        if last_dist is not None:
            cur_path = substring(path, last_dist, dist)
            if isinstance(cur_path, Point):
                cur_path = LineString((cur_path, cur_path))
            add_edge(last_projected_stop_id, node_id, geometry=cur_path, transport=True)
        last_projected_stop_id = node_id

        node_id += 1
        add_node("platform", platform.x, platform.y)
        add_edge(node_id - 1, node_id)
        add_edge(node_id, node_id - 1)
        node_id += 1
        return dist, last_projected_stop_id, node_id

    if not path:
        warnings.warn("В одном из маршрутов нет пути, пока не работаю с этим", UserWarning)
        return None
    path = LineString(path)

    # Если нет платформ
    if not platforms:
        if not stops:  # Строим маршрут по пути начало - конец
            platforms = [offset_point(path.interpolate(0), 1, 7), offset_point(path.interpolate(path.length), 1, 7)]
        else:  # Если есть только остановки - превращаем их в платформы
            platforms = [offset_point(Point(stop), 1, 7) for stop in stops]
            stops = None

    stops = [] if not stops else stops

    # Если остановок больше чем платформ - найти остановки без платформ и добавить новые платформы
    if len(stops) > len(platforms):
        stop_tree = cKDTree(stops)
        _, indices = stop_tree.query(platforms)
        connection = [(platforms[platform], stop) for platform, stop in enumerate(indices)]
        connection += [(-1, stop) for stop in set(range(len(stops))) ^ set(indices)]
        connection.sort(key=lambda x: x[1])
        direction = offset_direction(Point(platforms[len(platforms) // 2]))
        stops_to_platforms = {
            stop: offset_point(Point(stops[stop]), direction, 7) for stop in set(range(len(stops))) ^ set(indices)
        }
        platforms = [coord if (coord != -1) else (stops_to_platforms.get(ind)) for coord, ind in connection]
        stops = []

    # Если получилось только одна платформа
    if len(platforms) == 1:
        platform = Point(platforms[0])
        dist = path.project(platform)
        if dist in (path.length, 0):  # Если платформа является конечной
            platforms = [offset_point(path.interpolate(0), 1, 7), offset_point(path.interpolate(path.length), 1, 7)]
        else:  # Если платформа не является конечной
            platforms = [
                offset_point(path.interpolate(0), 1, 7),
                platform,
                offset_point(path.interpolate(path.length), 1, 7),
            ]
    platforms = [Point(coords) for coords in platforms]
    if len(platforms) >= len(stops):
        for platform in platforms:
            if not last_dist:
                last_dist, last_projected_stop_id, node_id = process_platform(platform)
                if last_dist > path.length / 2:
                    path = path.reverse()
                    last_dist = path.project(platform)
            else:
                last_dist, last_projected_stop_id, node_id = process_platform(platform)

    to_return = pd.DataFrame(graph_data)
    return to_return


def parse_overpass_to_edgenode(loc, crs) -> pd.DataFrame | None:
    loc_id = loc.name
    transport_type = loc.transport_type
    parsed_geometry = parse_overpass_route_response(loc, crs)
    edgenode = geometry_to_graph_edge_node_df(parsed_geometry, transport_type, loc_id)
    return edgenode



# pylint: disable=unused-import
from .modules.downloaders import get_boundary
from .modules.drive_walk_builder import get_drive_graph, get_walk_graph
from .modules.graph_transformer import graph_to_gdf
from .modules.intermodal_builder import get_intermodal_graph
from .modules.matrix.matrix_builder import get_adj_matrix_gdf_to_gdf, get_closest_nodes
from .modules.pt_walk_joiner import join_pt_walk_graph
from .modules.public_transport_builder import get_all_public_transport_graph, get_single_public_transport_graph



import multiprocessing

import geopandas as gpd
import networkx as nx
import pandas as pd
from shapely import MultiPolygon, Polygon
from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map, thread_map

from iduedu import config
from iduedu.enums.pt_enums import PublicTrasport
from iduedu.modules.routes_parser import parse_overpass_to_edgenode
from iduedu.utils.utils import clip_nx_graph, estimate_crs_for_bounds

from .downloaders import (
    get_boundary,
    get_routes_by_poly,
)

logger = config.logger


def _graph_data_to_nx(graph_df, keep_geometry: bool = True) -> nx.DiGraph:
    platforms = graph_df[graph_df["type"] == "platform"]
    platforms = platforms.groupby("point", as_index=False).agg({"node_id": list, "route": list})
    platforms["type"] = "platform"

    stops = graph_df[(graph_df["type"] != "platform") & (graph_df["u"].isna())][["point", "node_id", "route", "type"]]
    stops = stops.groupby(["point", "route", "type"], as_index=False).agg({"node_id": list})
    stops["route"] = stops["route"].apply(lambda x: [x])

    all_nodes = pd.concat([platforms, stops], ignore_index=True).reset_index(drop=True)
    mapping = {}
    for i, row in all_nodes.iterrows():
        index = row.name
        node_id_list = row["node_id"]
        for node_id in node_id_list:
            mapping[node_id] = int(index)

    def replace_with_mapping(value):
        return mapping.get(value)

    # Применение функции замены к каждому столбцу DataFrame
    graph_df["u"] = graph_df["u"].apply(replace_with_mapping)
    graph_df["v"] = graph_df["v"].apply(replace_with_mapping)
    graph_df["node_id"] = graph_df["node_id"].apply(replace_with_mapping)

    edges = graph_df[~graph_df["u"].isna()][["route", "type", "u", "v", "geometry"]].copy()

    def calc_len_time(row):
        if row.type == "boarding":
            return 0, 0
        length = round(row.geometry.length, 3)
        return length, round(length / PublicTrasport[row.type.upper()].avg_speed, 3)

    edges[["length_meter", "time_min"]] = edges.apply(
        calc_len_time,
        axis=1,
        result_type="expand",
    )
    graph = nx.DiGraph()
    for i, node in all_nodes.iterrows():
        route = list(set(node["route"]))
        if len(route) == 1:
            route = route[0]
        graph.add_node(i, x=node["point"][0], y=node["point"][1], type=node["type"], route=route)
    for i, edge in edges.iterrows():
        graph.add_edge(
            edge["u"],
            edge["v"],
            route=edge["route"],
            type=edge["type"],
            geometry=edge["geometry"] if keep_geometry else None,
            length_meter=edge["length_meter"],
            time_min=edge["time_min"],
        )

    return graph


def _get_multi_routes_by_poly(args):
    polygon, transport = args
    return get_routes_by_poly(polygon, transport)


def _multi_process_row(args):
    row, local_crs = args
    return parse_overpass_to_edgenode(row, local_crs)


def get_single_public_transport_graph(
    public_transport_type: str | PublicTrasport,
    osm_id: int | None = None,
    territory_name: str | None = None,
    polygon: Polygon | MultiPolygon | None = None,
    clip_by_bounds: bool = False,
    keep_geometry: bool = True,
):
    """
    Generate a graph of a specific type of public transport network within a given territory or polygon.
    The graph can optionally be clipped by the specified bounds and retain original geometries.

    Parameters
    ----------
    public_transport_type : str | PublicTransport
        Type of public transport to generate the graph for. Examples include "bus", "tram", "subway", etc.
    osm_id : int, optional
        OpenStreetMap ID of the territory. Either this or `territory_name` must be provided.
    territory_name : str, optional
        Name of the territory to generate the transport network for. Either this or `osm_id` must be provided.
    polygon : Polygon | MultiPolygon, optional
        A custom polygon or MultiPolygon to define the area for the transport network. Must be in CRS 4326.
    clip_by_bounds : bool, optional
        If True, clips the resulting graph to the bounds of the provided polygon. Defaults to False.
    keep_geometry : bool, optional
        If True, retains the original geometry of the transport routes. Defaults to True.

    Returns
    -------
    networkx.Graph
        A public transport network graph for the specified transport type and territory or polygon.

    Raises
    ------
    ValueError
        If no valid `osm_id`, `territory_name`, or `polygon` is provided.

    Warnings
    --------
    Logs a warning if no public transport routes are found in the specified area.

    Examples
    --------
    >>> bus_graph = get_single_public_transport_graph(public_transport_type="bus", osm_id=1114252)
    >>> tram_graph = get_single_public_transport_graph(territory_name="Санкт-Петербург", public_transport_type="tram")
    >>> metro_graph = get_single_public_transport_graph(polygon=poly, public_transport_type="bus", clip_by_bounds=True)

    Notes
    -----
    The function uses OpenStreetMap data and processes public transport routes in parallel.
    The CRS for the graph is estimated based on the bounds of the provided/downloaded polygon, stored in G.graph['crs'].
    """

    polygon = get_boundary(osm_id, territory_name, polygon)

    overpass_data = get_routes_by_poly(polygon, public_transport_type)
    if overpass_data.shape[0] == 0:
        logger.warning("No routes found for public transport.")
        return nx.Graph()
    local_crs = estimate_crs_for_bounds(*polygon.bounds).to_epsg()

    if not config.enable_tqdm_bar:
        logger.debug("Parsing routes")

    if overpass_data.shape[0] > 100:
        # Если много маршрутов - обрабатываем в параллели
        n_cpus = multiprocessing.cpu_count()
        rows = [(row, local_crs) for _, row in overpass_data.iterrows()]
        edgenode_for_routes = process_map(
            _multi_process_row, rows, desc="Parsing routes", chunksize=1, disable=not config.enable_tqdm_bar
        )
    else:
        tqdm.pandas(desc="Parsing routes data", disable=not config.enable_tqdm_bar)
        edgenode_for_routes = overpass_data.progress_apply(
            lambda x: parse_overpass_to_edgenode(x, local_crs), axis=1
        ).tolist()
    graph_df = pd.concat(edgenode_for_routes, ignore_index=True)

    to_return = _graph_data_to_nx(graph_df, keep_geometry=keep_geometry)
    to_return.graph["crs"] = local_crs
    to_return.graph["type"] = public_transport_type

    if clip_by_bounds:
        polygon = gpd.GeoSeries([polygon], crs=4326).to_crs(local_crs).union_all()
        return clip_nx_graph(to_return, polygon)

    logger.debug("Done!")
    return to_return


def get_all_public_transport_graph(
    osm_id: int | None = None,
    territory_name: str | None = None,
    polygon: Polygon | MultiPolygon | None = None,
    clip_by_bounds: bool = False,
    keep_geometry: bool = True,
    transport_types: list[PublicTrasport] = None,
) -> nx.Graph:
    """
    Generate a public transport network graph that includes all types of transport within a specified territory.
    The graph can optionally be clipped by bounds and retain original geometries.

    Parameters
    ----------
    osm_id : int, optional
        OpenStreetMap ID of the territory. Either this or `territory_name` must be provided.
    territory_name : str, optional
        Name of the territory to generate the public transport network for. Either this or `osm_id` must be provided.
    polygon : Polygon | MultiPolygon, optional
        A custom polygon or MultiPolygon to define the area for the transport network. Must be in CRS 4326.
    clip_by_bounds : bool, optional
        If True, clips the resulting graph to the bounds of the provided polygon. Defaults to False.
    keep_geometry : bool, optional
        If True, retains the original geometry of the transport routes. Defaults to True.
    transport_types: list[PublicTransport], optional
        By default `[PublicTrasport.TRAM, PublicTrasport.BUS, PublicTrasport.TROLLEYBUS, PublicTrasport.SUBWAY]`,
        can be any combination of PublicTransport Enums.

    Returns
    -------
    networkx.Graph
        A public transport network graph for all transport types (e.g., bus, tram, subway) in the specified territory.

    Raises
    ------
    ValueError
        If no valid `osm_id`, `territory_name`, or `polygon` is provided.

    Warnings
    --------
    Logs a warning if no public transport routes are found in the specified area.

    Examples
    --------
    >>> all_transport_graph = get_all_public_transport_graph(osm_id=1114252)
    >>> all_transport_graph = get_all_public_transport_graph(territory_name="Санкт-Петербург", clip_by_bounds=True)
    >>> all_transport_graph = get_all_public_transport_graph(polygon=some_polygon, keep_geometry=False)

    Notes
    -----
    This function retrieves routes for all public transport types using OpenStreetMap data.
    The CRS for the graph is estimated based on the bounds of the provided/downloaded polygon, stored in G.graph['crs'].
    """

    if transport_types is None:
        transport_types = [PublicTrasport.TRAM, PublicTrasport.BUS, PublicTrasport.TROLLEYBUS, PublicTrasport.SUBWAY]
    else:
        for transport_type in transport_types:
            if not isinstance(transport_type, PublicTrasport):
                raise ValueError(f"transport_type {transport_type} is not a valid transport type.")

    polygon = get_boundary(osm_id, territory_name, polygon)

    transports = [transport.value for transport in transport_types]
    args_list = [(polygon, transport) for transport in transports]

    if not config.enable_tqdm_bar:
        logger.debug("Downloading pt routes")

    overpass_data = pd.concat(
        thread_map(
            _get_multi_routes_by_poly,
            args_list,
            desc="Downloading public transport routes from OSM",
            disable=not config.enable_tqdm_bar,
        ),
        ignore_index=True,
    ).reset_index(drop=True)

    if overpass_data.shape[0] == 0:
        logger.warning("No routes found for public transport.")
        return nx.Graph()

    local_crs = estimate_crs_for_bounds(*polygon.bounds).to_epsg()

    if not config.enable_tqdm_bar:
        logger.debug("Parsing public transport routes")
    if overpass_data.shape[0] > 100:
        # Если много маршрутов - обрабатываем в параллели
        rows = [(row, local_crs) for _, row in overpass_data.iterrows()]
        edgenode_for_routes = process_map(
            _multi_process_row,
            rows,
            desc="Parsing public transport routes",
            chunksize=1,
            disable=not config.enable_tqdm_bar,
        )
    else:
        tqdm.pandas(desc="Parsing public transport routes", disable=not config.enable_tqdm_bar)
        edgenode_for_routes = overpass_data.progress_apply(
            lambda x: parse_overpass_to_edgenode(x, local_crs), axis=1
        ).tolist()

    graph_df = pd.concat(edgenode_for_routes, ignore_index=True)
    to_return = _graph_data_to_nx(graph_df, keep_geometry=keep_geometry)
    to_return.graph["crs"] = local_crs
    to_return.graph["type"] = "public_trasport"

    if clip_by_bounds:
        polygon = gpd.GeoSeries([polygon], crs=4326).to_crs(local_crs).union_all()
        to_return = clip_nx_graph(to_return, polygon)

    logger.debug("Done!")
    return to_return



from enum import Enum
from typing import Literal


class RegistrationStatus(Enum):
    """
    Enum for registration status of the road.
    """

    FEDERAL = 1
    REGIONAL = 2
    LOCAL = 3


class HighwayType(Enum):
    """
    Enum of highway types. Properties contain registration status & max speeds
    """

    MOTORWAY = "motorway"
    TRUNK = "trunk"
    PRIMARY = "primary"
    SECONDARY = "secondary"
    TERTIARY = "tertiary"
    UNCLASSIFIED = "unclassified"
    RESIDENTIAL = "residential"
    MOTORWAY_LINK = "motorway_link"
    TRUNK_LINK = "trunk_link"
    PRIMARY_LINK = "primary_link"
    SECONDARY_LINK = "secondary_link"
    TERTIARY_LINK = "tertiary_link"
    LIVING_STREET = "living_street"

    @property
    def reg_status(self) -> Literal[1, 2, 3]:
        reg_status = {
            HighwayType.MOTORWAY: RegistrationStatus.FEDERAL,
            HighwayType.TRUNK: RegistrationStatus.FEDERAL,
            HighwayType.PRIMARY: RegistrationStatus.REGIONAL,
            HighwayType.SECONDARY: RegistrationStatus.REGIONAL,
            HighwayType.TERTIARY: RegistrationStatus.LOCAL,
            HighwayType.UNCLASSIFIED: RegistrationStatus.LOCAL,
            HighwayType.RESIDENTIAL: RegistrationStatus.LOCAL,
            HighwayType.MOTORWAY_LINK: RegistrationStatus.FEDERAL,
            HighwayType.TRUNK_LINK: RegistrationStatus.FEDERAL,
            HighwayType.PRIMARY_LINK: RegistrationStatus.REGIONAL,
            HighwayType.SECONDARY_LINK: RegistrationStatus.REGIONAL,
            HighwayType.TERTIARY_LINK: RegistrationStatus.LOCAL,
            HighwayType.LIVING_STREET: RegistrationStatus.LOCAL,
        }
        return reg_status[self].value

    @property
    def max_speed(self) -> float:
        """
        Average speed in m/min.
        """
        speeds = {  # km/h
            HighwayType.MOTORWAY: 110,
            HighwayType.TRUNK: 90,
            HighwayType.PRIMARY: 60,
            HighwayType.SECONDARY: 60,
            HighwayType.TERTIARY: 60,
            HighwayType.UNCLASSIFIED: 40,
            HighwayType.RESIDENTIAL: 40,
            HighwayType.MOTORWAY_LINK: 90,
            HighwayType.TRUNK_LINK: 90,
            HighwayType.PRIMARY_LINK: 60,
            HighwayType.SECONDARY_LINK: 60,
            HighwayType.TERTIARY_LINK: 60,
            HighwayType.LIVING_STREET: 20,
        }
        return speeds[self] * 1000 / 60  # metres per minute



from enum import Enum


class PublicTrasport(Enum):
    """
    Enumeration class for edge types in graphs.
    """

    SUBWAY = "subway"
    BUS = "bus"
    TRAM = "tram"
    TROLLEYBUS = "trolleybus"
    TRAIN = "train"

    @property
    def avg_speed(self) -> float:
        """
        Average speed in m/min.
        """
        speeds = {  # km/h
            PublicTrasport.SUBWAY: 40,
            PublicTrasport.BUS: 20,
            PublicTrasport.TRAM: 20,
            PublicTrasport.TROLLEYBUS: 18,
            PublicTrasport.TRAIN: 60,
        }
        return speeds[self] * 1000 / 60



import osmnx as ox
import pandas as pd
import requests
from shapely import LineString, MultiPolygon, Polygon, unary_union
from shapely.ops import polygonize

from iduedu import config

logger = config.logger


class RequestError(RuntimeError):
    """
    Basic error for eny requests problems
    """

    def __init__(self, message, status_code=None, reason=None, response_text=None, response_content=None):
        super().__init__(message)
        self.status_code = status_code
        self.reason = reason
        self.response_text = response_text
        self.response_content = response_content

    def __str__(self):
        if self.status_code == 400:
            return (
                f"{super().__str__()} (status: {self.status_code}, reason: {self.reason}). "
                f"Make sure provided polygon is in CRS 4326."
            )
        return f"{super().__str__()} (status: {self.status_code}, reason: {self.reason})."


def get_boundary_by_osm_id(osm_id) -> MultiPolygon | Polygon:
    overpass_query = f"""
                [out:json];
                        (
                            relation({osm_id});
                        );
                out geom;
                """
    logger.debug(f"Downloading territory bounds with osm_id <{osm_id}> ...")
    result = requests.get(config.overpass_url, params={"data": overpass_query}, timeout=config.timeout)
    if result.status_code == 200:
        json_result = result.json()
        geoms = []
        for element in json_result["elements"]:
            geometries_inners = []
            geometries_outers = []
            for member in element["members"]:
                if "geometry" in member:
                    if member["role"] == "outer":
                        geometries_outers.append(
                            LineString([(coords["lon"], coords["lat"]) for coords in member["geometry"]])
                        )
                    if member["role"] == "inner":
                        geometries_inners.append(
                            LineString([(coords["lon"], coords["lat"]) for coords in member["geometry"]])
                        )
            outer_poly = unary_union(list(polygonize(geometries_outers)))
            inner_poly = unary_union(list(polygonize(geometries_inners)))
            geoms.append(outer_poly.difference(inner_poly))
        return unary_union(geoms)
    raise RequestError(
        message=f"Request failed with status code {result.status_code}, reason: {result.reason}",
        status_code=result.status_code,
        reason=result.reason,
        response_text=result.text,
        response_content=result.content,
    )


def get_boundary_by_name(territory_name: str) -> Polygon | MultiPolygon:
    logger.debug(f"Downloading territory bounds with name <{territory_name}> ...")
    place = ox.geocode_to_gdf(territory_name)
    return unary_union(place.geometry)


def get_boundary(
    osm_id: int | None = None, territory_name: str | None = None, polygon: Polygon | MultiPolygon | None = None
) -> Polygon:
    """
    Retrieve the boundary polygon for a given territory, either by OSM ID, territory name, or an existing polygon.
    If a MultiPolygon is provided, it will be converted to a convex hull.

    Parameters
    ----------
    osm_id : int, optional
        OpenStreetMap ID of the territory to retrieve the boundary for.
        Either this or `territory_name` must be provided.
    territory_name : str, optional
        Name of the territory to retrieve the boundary for. Either this or `osm_id` must be provided.
    polygon : Polygon | MultiPolygon, optional
        A custom polygon or MultiPolygon to use instead of querying by `osm_id` or `territory_name`.

    Returns
    -------
    Polygon
        The boundary polygon for the specified territory.
        If a MultiPolygon was provided, it will return the convex hull.

    Raises
    ------
    ValueError
        If neither `osm_id`, `territory_name`, nor `polygon` are provided.

    Examples
    --------
    >>> boundary = get_boundary(osm_id=123456)
    >>> boundary = get_boundary(territory_name="New York")
    >>> boundary = get_boundary(polygon=some_polygon)
    """

    if osm_id is None and territory_name is None and polygon is None:
        raise ValueError("Either osm_id or name or polygon must be specified")
    if osm_id:
        polygon: Polygon = get_boundary_by_osm_id(osm_id)
    elif territory_name:
        polygon: Polygon = get_boundary_by_name(territory_name)

    if isinstance(polygon, MultiPolygon):
        polygon: Polygon = polygon.convex_hull

    return polygon


def get_routes_by_poly(polygon: Polygon, public_transport_type: str) -> pd.DataFrame:
    polygon_coords = " ".join(f"{y} {x}" for x, y in polygon.exterior.coords[:-1])
    overpass_query = f"""
        [out:json];
                (
                    relation(poly:"{polygon_coords}")['route'='{public_transport_type}'];
                );
        out geom;
        """
    logger.debug(f"Downloading routes from OSM with type <{public_transport_type}> ...")
    result = requests.post(config.overpass_url, data={"data": overpass_query}, timeout=config.timeout)
    if result.status_code == 200:
        json_result = result.json()["elements"]
        data = pd.DataFrame(json_result)
        data["transport_type"] = public_transport_type
        return data
    raise RequestError(
        message=f"Request failed with status code {result.status_code}, reason: {result.reason}",
        status_code=result.status_code,
        reason=result.reason,
        response_text=result.text,
        response_content=result.content,
    )



import geopandas as gpd
import networkx as nx
import pandas as pd
from loguru import logger
from shapely import LineString
from shapely.geometry.point import Point


def _edges_to_gdf(graph: nx.Graph, crs) -> gpd.GeoDataFrame:
    """
    Converts nx graph to gpd.GeoDataFrame as edges.
    """
    graph_df = pd.DataFrame(list(graph.edges(data=True)), columns=["u", "v", "data"])
    edge_data_expanded = pd.json_normalize(graph_df["data"])
    graph_df = pd.concat([graph_df.drop(columns=["data"]), edge_data_expanded], axis=1)
    graph_df = gpd.GeoDataFrame(graph_df, geometry="geometry", crs=crs).set_index(["u", "v"])
    graph_df["geometry"] = graph_df["geometry"].fillna(LineString())
    return graph_df


def _nodes_to_gdf(graph: nx.Graph, crs: int) -> gpd.GeoDataFrame:
    """
    Converts nx graph to gpd.GeoDataFrame as nodes.
    """

    ind, data = zip(*graph.nodes(data=True))
    node_geoms = (Point(d["x"], d["y"]) for d in data)
    gdf_nodes = gpd.GeoDataFrame(data, index=ind, crs=crs, geometry=list(node_geoms))

    return gdf_nodes


def _restore_edges_geom(nodes_gdf, edges_gdf) -> gpd.GeoDataFrame:
    edges_wout_geom = edges_gdf[edges_gdf["geometry"].is_empty].reset_index()
    edges_wout_geom["geometry"] = [
        LineString((s, e))
        for s, e in zip(
            nodes_gdf.loc[edges_wout_geom["u"], "geometry"], nodes_gdf.loc[edges_wout_geom["v"], "geometry"]
        )
    ]
    edges_wout_geom.set_index(["u", "v"], inplace=True)
    edges_gdf.update(edges_wout_geom)
    return edges_gdf


def graph_to_gdf(
    graph: nx.MultiDiGraph, edges: bool = True, nodes: bool = True, restore_edge_geom=False
) -> gpd.GeoDataFrame | None:
    """
    Converts nx graph to gpd.GeoDataFrame as edges.

    Parameters
    ----------
    graph : nx.MultiDiGraph
        The graph to convert.
    edges: bool, default to True
        Keep edges in GoeDataFrame.
    nodes: bool, default to True
        Keep nodes in GoeDataFrame.
    restore_edge_geom: bool, default to False
        if True, will try to restore edge geometry from nodes.
    Returns
    -------
    gpd.GeoDataFrame
        Graph representation in GeoDataFrame format
    """
    try:
        crs = graph.graph["crs"]
    except KeyError as exc:
        raise ValueError("Graph does not have crs attribute and no crs was provided") from exc
    if not edges and not nodes:
        logger.debug("Neither edges or nodes were selected, graph_to_gdf returning None")
        return None
    if nodes and not edges:
        nodes_gdf = _nodes_to_gdf(graph, crs)
        return nodes_gdf
    if not nodes and edges:
        edges_gdf = _edges_to_gdf(graph, crs)
        if restore_edge_geom:
            nodes_gdf = _nodes_to_gdf(graph, crs)
            edges_gdf = _restore_edges_geom(nodes_gdf, edges_gdf)
        return edges_gdf

    nodes_gdf = _nodes_to_gdf(graph, crs)
    edges_gdf = _edges_to_gdf(graph, crs)
    if restore_edge_geom:
        edges_gdf = _restore_edges_geom(nodes_gdf, edges_gdf)
    full_gdf = pd.concat([nodes_gdf, edges_gdf])
    return full_gdf



import concurrent.futures
import time

import networkx as nx
from shapely import MultiPolygon, Polygon

from iduedu import config
from iduedu.enums.pt_enums import PublicTrasport
from iduedu.modules.downloaders import get_boundary
from iduedu.modules.drive_walk_builder import get_walk_graph
from iduedu.modules.pt_walk_joiner import join_pt_walk_graph
from iduedu.modules.public_transport_builder import get_all_public_transport_graph

logger = config.logger


def get_intermodal_graph(
    osm_id: int | None = None,
    territory_name: str | None = None,
    polygon: Polygon | MultiPolygon | None = None,
    clip_by_bounds: bool = False,
    keep_routes_geom: bool = True,
    max_dist: float = 30,
    transport_types: list[PublicTrasport] = None,
) -> nx.Graph:
    """
    Generate an intermodal transport graph that combines public transport and pedestrian networks,
    with platforms serving as connection points between the two graphs.

    Parameters
    ----------
    osm_id : int, optional
        OpenStreetMap ID of the territory. Either this or `territory_name` must be provided.
    territory_name : str, optional
        Name of the territory to generate the intermodal transport network for.
        Either this or `osm_id` must be provided.
    polygon : Polygon | MultiPolygon, optional
        A custom polygon or MultiPolygon defining the area for the intermodal network. Must be in CRS 4326.
    clip_by_bounds : bool, optional
        If True, clips the public transport network to the bounds of the provided polygon. Defaults to False.
    keep_routes_geom : bool, optional
    max_dist : float, optional
        Maximum distance (in meters) to search for connections between platforms and pedestrian edges. Defaults to 30.
    transport_types: list[PublicTransport], optional
        By default `[PublicTrasport.TRAM, PublicTrasport.BUS, PublicTrasport.TROLLEYBUS, PublicTrasport.SUBWAY]`,
        can be any combination of PublicTransport Enums.
    Returns
    -------
    nx.Graph
        An intermodal network graph combining public transport and pedestrian routes, where public transport platforms
        are linked to nearby walking routes.

    Raises
    ------
    ValueError
        If no valid `osm_id`, `territory_name`, or `polygon` is provided.

    Warnings
    --------
    Logs a warning if the public transport graph is empty and only returns the pedestrian graph.

    Examples
    --------
    >>> intermodal_graph = get_intermodal_graph(osm_id=1114252, clip_by_bounds=True)
    >>> intermodal_graph = get_intermodal_graph(territory_name="Санкт-Петербург", polygon=some_polygon)

    Notes
    -----
    The function concurrently downloads and processes both the pedestrian and public transport graphs,
    then combines them using platforms as connection points.
    If the public transport graph is empty, only the pedestrian graph is returned.
    The CRS for the graph is estimated based on the bounds of the provided/downloaded polygon, stored in G.graph['crs'].
    """

    boundary = get_boundary(osm_id, territory_name, polygon)
    with concurrent.futures.ThreadPoolExecutor() as executor:
        walk_graph_future = executor.submit(get_walk_graph, polygon=boundary)
        logger.debug("Started downloading and parsing walk graph...")

        # Sleep to not get 429 to many requests
        time.sleep(0.5)
        pt_graph_future = executor.submit(
            get_all_public_transport_graph,
            polygon=boundary,
            clip_by_bounds=clip_by_bounds,
            keep_geometry=keep_routes_geom,
            transport_types=transport_types,
        )
        logger.debug("Started downloading and parsing public trasport graph...")

        pt_g = pt_graph_future.result()
        logger.debug("Public trasport graph done!")

        walk_g = walk_graph_future.result()
        logger.debug("Walk graph done!")

    if len(pt_g.nodes()) == 0:
        logger.warning("Public trasport graph is empty! Returning only walk graph.")
        return walk_g

    intermodal = join_pt_walk_graph(pt_g, walk_g, max_dist=max_dist)
    return intermodal



import re

import geopandas as gpd
import networkx as nx
import osmnx as ox
import pandas as pd
from shapely import MultiPolygon, Polygon, unary_union
from tqdm.auto import tqdm

from iduedu import config
from iduedu.enums.drive_enums import HighwayType
from iduedu.modules.downloaders import get_boundary
from iduedu.utils.utils import estimate_crs_for_bounds, remove_weakly_connected_nodes

logger = config.logger

BASE_FILTER = "['highway'~'" + "|".join([h.value for h in HighwayType]) + "']"


def highway_type_to_reg(highway_type) -> int:
    """
    Determine the reg_status based on highway type.
    """
    try:
        if isinstance(highway_type, list):
            reg_values = [HighwayType[ht.upper()].reg_status for ht in highway_type]
            return min(reg_values)
        return HighwayType[highway_type.upper()].reg_status
    except KeyError:
        return 3


def determine_reg(name_roads, highway_type=None) -> int:
    """
    Determine the reg_status based on road_name.
    """

    if isinstance(name_roads, list):
        for item in name_roads:
            if re.match(r"^[МАР]", str(item)):
                return 1
            if re.match(r"^\d.*[A-Za-zА-Яа-я]", str(item)):
                return 2
        return 3
    if pd.isna(name_roads):
        # Выставление значения по типу дороги, если значение NaN
        if highway_type:
            return highway_type_to_reg(highway_type)
        return 3
    if re.match(r"^[МАР]", str(name_roads)):
        return 1
    if re.match(r"^\d.*[A-Za-zА-Яа-я]", str(name_roads)):
        return 2
    return 3


def get_max_speed(highway_types) -> float:
    """
    Determine the speed based on road_name.
    """
    # Проверяем, является ли highway_types списком.
    try:
        if isinstance(highway_types, list):
            max_speeds = []
            for ht in highway_types:
                try:
                    highway_enum = HighwayType[ht.upper()]
                    max_speeds.append(highway_enum.max_speed)
                except KeyError:
                    logger.debug(f"{ht} not found in HighwayType enum, skipping.")
            if max_speeds:
                return max(max_speeds)
            logger.debug("No valid highway types provided, returning 40 km/h.")
            return 40 * 1000 / 60
        return HighwayType[highway_types.upper()].max_speed
    except KeyError:
        return 40 * 1000 / 60


def get_drive_graph_by_poly(
    polygon: Polygon | MultiPolygon, additional_edgedata=None, road_filter: str = None
) -> nx.MultiDiGraph:
    if additional_edgedata is None:
        additional_edgedata = []
    if not road_filter:
        road_filter = BASE_FILTER
    if isinstance(polygon, MultiPolygon):
        polygon = unary_union(polygon)
        if isinstance(polygon, MultiPolygon):
            polygon = polygon.convex_hull
    logger.info("Downloading drive graph from OSM, it may take a while for large territory ...")
    graph = ox.graph_from_polygon(
        polygon,
        network_type="drive",
        custom_filter=road_filter,
        truncate_by_edge=False,
    )
    local_crs = estimate_crs_for_bounds(*polygon.bounds).to_epsg()

    nodes, edges = ox.graph_to_gdfs(graph)
    edges: gpd.GeoDataFrame
    edges.reset_index(inplace=True)
    if "ref" not in edges.columns:
        edges["ref"] = pd.NA
    edges["reg"] = edges.apply(lambda row: determine_reg(row["ref"], row["highway"]), axis=1)

    nodes.to_crs(local_crs, inplace=True)
    nodes.geometry = nodes.geometry.set_precision(0.00001)
    nodes[["x", "y"]] = nodes.apply(lambda row: (row.geometry.x, row.geometry.y), axis=1, result_type="expand")
    edges.to_crs(local_crs, inplace=True)
    edges.geometry = edges.geometry.set_precision(0.00001)
    edges["maxspeed"] = edges["highway"].apply(get_max_speed)

    edges[["length_meter", "time_min"]] = edges.apply(
        lambda row: (round(row.geometry.length, 3), round(row.geometry.length / row.maxspeed, 3)),
        axis=1,
        result_type="expand",
    )
    edgesdata = ["u", "v", "key", "length_meter", "time_min", "geometry"] + additional_edgedata

    edges = edges[edgesdata]

    edges.set_index(["u", "v", "key"], inplace=True)
    graph = ox.graph_from_gdfs(nodes, edges)
    graph = remove_weakly_connected_nodes(graph)
    mapping = {old_label: new_label for new_label, old_label in enumerate(graph.nodes())}
    graph = nx.relabel_nodes(graph, mapping)
    graph.graph["crs"] = local_crs
    graph.graph["type"] = "drive"
    logger.debug("Done!")
    return graph


def get_drive_graph(
    osm_id: int | None = None,
    territory_name: str | None = None,
    polygon: Polygon | MultiPolygon | None = None,
    additional_edgedata=None,
):
    """
    Generate a road network graph for driving within the specified territory or polygon.
    Optionally, include additional edge data such as highway type, max speed, registration status and road name.

    Parameters
    ----------
    osm_id : int, optional
        OpenStreetMap ID of the territory to build the graph for. Either this or `territory_name` must be provided.
    territory_name : str, optional
        Name of the territory to build the graph for. Either this or `osm_id` must be provided.
    polygon : Polygon | MultiPolygon, optional
        A custom polygon or MultiPolygon to define the area for the road network. Must be in CRS 4326.
    additional_edgedata : list[str], optional
        List of additional edge data attributes to include in the graph. Possible values include
        ['highway', 'maxspeed', 'reg', 'ref', 'name'] or any other, that exist in OSM. Defaults to None.

    Returns
    -------
    networkx.Graph
        A road network graph for the specified territory or polygon, with optional additional edge data.

    Examples
    --------
    >>> drive_graph = get_drive_graph(osm_id=1114252)
    >>> drive_graph = get_drive_graph(territory_name="Санкт-Петербург", additional_edgedata=['highway', 'maxspeed'])
    >>> drive_graph = get_drive_graph(polygon=some_polygon, additional_edgedata=['name', 'ref'])

    Notes
    -----
    Road speeds are defined in `iduedu.enums.drive_enums.py`.
    The CRS for the graph is estimated based on the bounds of the provided/downloaded polygon, stored in G.graph['crs'].
    """

    polygon = get_boundary(osm_id, territory_name, polygon)

    return get_drive_graph_by_poly(polygon, additional_edgedata=additional_edgedata)


def get_walk_graph(
    osm_id: int | None = None,
    territory_name: str | None = None,
    polygon: Polygon | MultiPolygon | None = None,
    walk_speed: float = 5 * 1000 / 60,
):
    """
    Generate a pedestrian road network graph within the specified territory or polygon.
    The graph's edges includes calculated walking times based on the specified walking speed.

    Parameters
    ----------
    osm_id : int, optional
        OpenStreetMap ID of the territory to build the walking graph for.
        Either this or `territory_name` must be provided.
    territory_name : str, optional
        Name of the territory to build the walking graph for. Either this or `osm_id` must be provided.
    polygon : Polygon | MultiPolygon, optional
        A custom polygon or MultiPolygon to define the area for the pedestrian network. Must be in CRS 4326.
    walk_speed : float, optional
        Walking speed in meters per minute. Defaults to 5 km/h (approximately 83.33 meters per minute).

    Returns
    -------
    networkx.Graph
        A pedestrian road network graph with edge lengths and walking times for the specified territory or polygon.

    Examples
    --------
    >>> walk_graph = get_walk_graph(osm_id=1114252)
    >>> walk_graph = get_walk_graph(territory_name="Санкт-Петербург", walk_speed=5)
    >>> walk_graph = get_walk_graph(polygon=some_polygon)

    Notes
    -----
    The CRS for the graph is estimated based on the bounds of the provided/downloaded polygon, stored in G.graph['crs'].
    """

    polygon = get_boundary(osm_id, territory_name, polygon)

    logger.info("Downloading walk graph from OSM, it may take a while for large territory ...")
    graph = ox.graph_from_polygon(polygon, network_type="walk", truncate_by_edge=False, simplify=True)
    local_crs = estimate_crs_for_bounds(*polygon.bounds).to_epsg()

    nodes, edges = ox.graph_to_gdfs(graph)
    nodes.to_crs(local_crs, inplace=True)
    nodes.geometry = nodes.geometry.set_precision(0.00001)
    nodes[["x", "y"]] = nodes.apply(lambda row: (row.geometry.x, row.geometry.y), axis=1, result_type="expand")
    nodes = nodes[["x", "y", "geometry"]]
    edges.reset_index(inplace=True)
    edges.to_crs(local_crs, inplace=True)
    edges.geometry = edges.geometry.set_precision(0.00001)
    tqdm.pandas(desc="Calculating the weights of the walk graph", disable=not config.enable_tqdm_bar)
    edges[["length_meter", "time_min"]] = edges.progress_apply(
        lambda row: (round(row.geometry.length, 3), round(row.geometry.length / walk_speed, 3)),
        axis=1,
        result_type="expand",
    )
    edges["type"] = "walk"
    edges = edges[
        [
            "u",
            "v",
            "key",
            "length_meter",
            "time_min",
            "type",
            "geometry",
        ]
    ]
    edges.set_index(["u", "v", "key"], inplace=True)
    graph = ox.graph_from_gdfs(nodes, edges)
    graph = remove_weakly_connected_nodes(graph)
    mapping = {old_label: new_label for new_label, old_label in enumerate(graph.nodes())}
    graph = nx.relabel_nodes(graph, mapping)
    graph.graph["crs"] = local_crs
    graph.graph["walk_speed"] = walk_speed
    graph.graph["type"] = "walk"
    logger.debug("Done!")
    return graph



from heapq import heappop, heappush
from typing import Any, Iterable, List, Literal, Tuple

import geopandas as gpd
import networkx as nx
import numba as nb
import numpy as np
import pandas as pd
from numpy.core.multiarray import ndarray
from pyproj import CRS
from pyproj.exceptions import CRSError
from scipy.spatial import KDTree

from iduedu import config
from iduedu.modules.matrix.numba_csr_matrix import UI32CSRMatrix
from iduedu.utils.utils import remove_weakly_connected_nodes

logger = config.logger


@nb.njit(
    [
        nb.types.Array(nb.int32, 2, "C")(
            UI32CSRMatrix.class_type.instance_type,
            nb.types.Array(nb.int32, 1, "C"),
            nb.types.Array(nb.int32, 1, "C"),
            nb.int32,
        )
    ],
    cache=True,
    parallel=True,
)
def dijkstra_numba_parallel(
    numba_matrix: UI32CSRMatrix, sources: np.array, targets: np.array, cutoff: np.int32
):  # pragma: no cover
    distance_matrix = np.full((len(sources), len(targets)), -1, dtype=np.int32)

    for i in nb.prange(len(sources)):
        source = sources[i]
        distances = np.full(numba_matrix.tot_rows, -1, dtype=np.int32)
        seen = np.copy(distances)
        fringe = []
        seen[source] = np.int32(0)
        fringe.append((np.int32(0), np.int32(source)))
        while fringe:
            d, v = heappop(fringe)
            if distances[v] != -1:
                continue
            distances[v] = d
            for u, weight in zip(numba_matrix.get_nnz(v), numba_matrix.get_row(v)):
                uv_dist = np.int32(d + weight)
                if uv_dist > cutoff:
                    continue
                if seen[u] == -1 or uv_dist < seen[u]:
                    seen[u] = uv_dist
                    heappush(fringe, (uv_dist, np.int32(u)))

        distance_matrix[i, :] = distances[targets]
    return distance_matrix


def get_closest_nodes(
    gdf_from: gpd.GeoDataFrame, to_nx_graph: nx.Graph
) -> tuple[list[Any], float | ndarray | Iterable | int]:
    """
    Find the closest nodes in a NetworkX graph for each geometry in a GeoDataFrame.

    For each geometry in the input GeoDataFrame, this function identifies the nearest node in the provided graph
    based on Euclidean distance. The CRS of the GeoDataFrame is automatically aligned with the graph's CRS
    if they differ.

    Parameters
    ----------
    gdf_from : gpd.GeoDataFrame
        A GeoDataFrame containing the points for which the closest graph nodes are to be found.
    to_nx_graph : nx.Graph
        A NetworkX graph with geographic data, where each node has 'x' and 'y' attributes representing
        its coordinates in the graph's CRS.

    Returns
    -------
    tuple[list, list]
         returns a tuple with two lists where the first is the closest node ID and the second
          are the corresponding distances.

    Examples
    --------
    >>> closest_nodes = get_closest_nodes(points_gdf, graph)
    >>> closest_nodes_with_dist = get_closest_nodes(points_gdf, graph, return_dist=True)
    """
    try:
        graph_crs = CRS.from_epsg(to_nx_graph.graph["crs"])
    except CRSError:
        graph_crs = to_nx_graph.graph["crs"]

    if gdf_from.crs != graph_crs:
        gdf_from = gdf_from.to_crs(graph_crs)

    nodes_with_data = list(to_nx_graph.nodes(data=True))
    coordinates = [(data["x"], data["y"]) for node, data in nodes_with_data]
    tree = KDTree(coordinates)
    target_coord = [(p.x, p.y) for p in gdf_from.representative_point()]
    distances, indices = tree.query(target_coord)
    nearest_nodes = [nodes_with_data[idx][0] for idx in indices]

    return nearest_nodes, distances


def get_adj_matrix_gdf_to_gdf(
    gdf_from: gpd.GeoDataFrame,
    gdf_to: gpd.GeoDataFrame,
    nx_graph: nx.Graph,
    weight: Literal["length_meter", "time_min"] = "length_meter",
    dtype: np.dtype = np.float16,
    add_dist_tofrom_node=True,
    threshold: int = None,
    max_workers: int = None,
) -> pd.DataFrame:
    """
    Compute an adjacency matrix representing the shortest path distances between two sets of points (GeoDataFrames)
    based on a provided graph. Distances are calculated using the specified edge weight attribute.

    Parameters
    ----------
    gdf_from : gpd.GeoDataFrame
        The GeoDataFrame containing the origin points for the distance matrix calculation.
    gdf_to : gpd.GeoDataFrame
        The GeoDataFrame containing the destination points for the distance matrix calculation.
    nx_graph : nx.Graph
        A NetworkX graph with geographic data where each edge has the specified `weight` attribute (e.g., 'length_meter').
    weight : {"length_meter", "time_min"}, optional
        The edge attribute to use for calculating the shortest paths. Defaults to 'length_meter'.
    dtype : np.dtype, optional
        The data type for the adjacency matrix. Defaults to np.float16, which can be changed to avoid precision loss.
    threshold : int, optional
        The maximum path distance to consider when calculating distances. Paths longer than this value are ignored.
        Default is None, which sets the threshold to the maximum integer value (essentially no threshold).
    max_workers : int, optional
        The maximum number of threads to use during computation. By default, uses all available CPU cores (os.cpu_count()).

    Returns
    -------
    pd.DataFrame
        A DataFrame representing the adjacency matrix where rows correspond to `gdf_from` and columns to `gdf_to`.
        Each value in the matrix represents the shortest path distance between the corresponding points,
        based on the provided graph. If no path exists or the path is beyond the threshold, the entry is set to np.inf.

    Notes
    -----
    - The function computes the closest graph nodes for both `gdf_from` and `gft_to` before calculating the matrix.
    - If `gdf_from` and `gft_to` are equal, the distance matrix will be square.
    - Ensure that all edges in the graph have the specified `weight` attribute; otherwise, the calculation may fail.

    Examples
    --------
    >>> adj_matrix = get_adj_matrix_gdf_to_gdf(origins_gdf, destinations_gdf, graph, weight='time_min', dtype=np.float32)
    """
    try:
        local_crs = nx_graph.graph["crs"]
    except KeyError as exc:
        raise ValueError("Graph does not have crs attribute") from exc

    try:
        gdf_from = gdf_from.to_crs(nx_graph.graph["crs"])
        gdf_to = gdf_to.to_crs(nx_graph.graph["crs"])
    except CRSError as e:
        raise CRSError(f"Graph crs ({local_crs}) has invalid format.") from e

    nx_graph = remove_weakly_connected_nodes(nx_graph)

    logger.debug("Preparing graph sparse matrix")
    transposed = False

    if gdf_from.equals(gdf_to):
        closest_nodes_from, dist_from = get_closest_nodes(gdf_from, nx_graph)
        closest_nodes_to, dist_to = closest_nodes_from, dist_from
        sparse_row_scipy = _get_sparse_row(nx_graph, weight)
    else:
        dif = len(gdf_from) - len(gdf_to)
        if dif <= 0:  # straight
            closest_nodes_from, dist_from = get_closest_nodes(gdf_from, nx_graph)
            closest_nodes_to, dist_to = get_closest_nodes(gdf_to, nx_graph)
            sparse_row_scipy = _get_sparse_row(nx_graph, weight)

        else:  # reversed
            transposed = True
            closest_nodes_from, dist_from = get_closest_nodes(gdf_to, nx_graph)
            closest_nodes_to, dist_to = get_closest_nodes(gdf_from, nx_graph)
            sparse_row_scipy = _get_sparse_row(nx_graph, weight).transpose().tocsc().tocsr()

    csr_matrix = UI32CSRMatrix(*_get_numba_matrix_attr(sparse_row_scipy))
    del sparse_row_scipy

    if max_workers is not None:
        nb.set_num_threads(max_workers)

    # *** Warm-up ***
    dijkstra_numba_parallel(
        numba_matrix=csr_matrix,
        sources=np.array(closest_nodes_to[:1], dtype=np.int32),
        targets=np.array(closest_nodes_from[:1], dtype=np.int32),
        cutoff=np.int32(0),
    )

    if threshold is None:
        threshold = np.iinfo(np.int32).max
    else:
        threshold = threshold * 100

    logger.debug("Starting the gdf-to-gdf matrix calculation")

    adj_matrix = dijkstra_numba_parallel(
        numba_matrix=csr_matrix,
        sources=np.array(closest_nodes_from, dtype=np.int32),
        targets=np.array(closest_nodes_to, dtype=np.int32),
        cutoff=np.int32(threshold),
    )
    if transposed:
        adj_matrix = adj_matrix.transpose()

    adj_matrix = pd.DataFrame(adj_matrix / 100, columns=gdf_to.index, index=gdf_from.index, dtype=dtype)
    if add_dist_tofrom_node:
        if weight == "time_min":
            speed = 5 * 1000 / 60
            dist_to = dist_to / speed
            dist_from = dist_from / speed
        dist_from_matrix = np.array(dist_from)[:, np.newaxis]
        dist_to_matrix = np.array(dist_to)[np.newaxis, :]
        mask = adj_matrix > 0
        additional_matrix = (dist_from_matrix + dist_to_matrix).astype(dtype)
        if transposed:
            additional_matrix = additional_matrix.transpose()
        adj_matrix[mask] += additional_matrix
    return adj_matrix.where(adj_matrix >= 0, np.inf)


def _get_sparse_row(nx_graph, weight):
    if nx_graph.is_multigraph():
        if nx_graph.is_directed():
            nx_graph = nx.DiGraph(nx_graph)
        else:
            nx_graph = nx.Graph(nx_graph)
    sparse_row_scipy = nx.to_scipy_sparse_array(nx_graph, weight=weight)
    sparse_row_scipy.data = np.round(sparse_row_scipy.data * 100).astype(np.uint32)
    return sparse_row_scipy


def _get_numba_matrix_attr(sparse_row_scipy):
    values = sparse_row_scipy.data.astype(np.uint32)
    col_index = sparse_row_scipy.indices.astype(np.uint32)
    row_index = sparse_row_scipy.indptr.astype(np.uint32)
    return values, col_index, row_index



import sys
from typing import Literal

from loguru import logger


class Config:
    """
    A configuration class to manage global settings for the application, such as Overpass API URL, timeouts, and logging options.

    Attributes
    ----------
    overpass_url : str
        URL for accessing the Overpass API. Defaults to "http://lz4.overpass-api.de/api/interpreter".
    timeout : int or None
        Timeout in seconds for API requests. If None, no timeout is applied.
    enable_tqdm_bar : bool
        Enables or disables progress bars (via tqdm). Defaults to True.
    logger : Logger
        Logging instance to handle application logging.

    Methods
    -------
    change_logger_lvl(lvl: Literal["TRACE", "DEBUG", "INFO", "WARN", "ERROR"])
        Changes the logging level to the specified value.
    set_overpass_url(url: str)
        Sets a new Overpass API URL.
    set_timeout(timeout: int)
        Sets the timeout for API requests.
    set_enable_tqdm(enable: bool)
        Enables or disables progress bars in the application.
    """

    def __init__(
        self,
        overpass_url="http://lz4.overpass-api.de/api/interpreter",
        timeout=None,
        enable_tqdm_bar=True,
    ):
        self.overpass_url = overpass_url
        self.timeout = timeout
        self.enable_tqdm_bar = enable_tqdm_bar
        self.logger = logger

    def change_logger_lvl(self, lvl: Literal["TRACE", "DEBUG", "INFO", "WARN", "ERROR"]):
        self.logger.remove()
        self.logger.add(sys.stderr, level=lvl)

    def set_overpass_url(self, url: str):
        self.overpass_url = url

    def set_timeout(self, timeout: int):
        self.timeout = timeout

    def set_enable_tqdm(self, enable: bool):
        self.enable_tqdm_bar = enable


config = Config()
config.change_logger_lvl("INFO")



import geopandas as gpd
import networkx as nx
import numpy as np
from loguru import logger
from pyproj import CRS
from pyproj.aoi import AreaOfInterest

# pylint: disable=no-name-in-module
from pyproj.database import query_utm_crs_info
from shapely import Point, Polygon


def clip_nx_graph(graph: nx.Graph, polygon: Polygon) -> nx.Graph:
    crs = graph.graph["crs"]
    points = gpd.GeoDataFrame(
        data=[{"id": p_id, "geometry": Point(data["x"], data["y"])} for p_id, data in graph.nodes(data=True)], crs=crs
    ).clip(polygon, True)
    clipped = graph.subgraph(points["id"].tolist())
    return clipped


def remove_weakly_connected_nodes(graph: nx.DiGraph) -> nx.DiGraph:
    graph = graph.copy()

    weakly_connected_components = list(nx.weakly_connected_components(graph))
    if len(weakly_connected_components) > 1:
        logger.warning(
            f"Found {len(weakly_connected_components)} disconnected subgraphs in the network. "
            f"These are isolated groups of nodes with no connections between them. "
            f"Size of components: {[len(c) for c in weakly_connected_components]}"
        )

    all_scc = sorted(nx.strongly_connected_components(graph), key=len)
    nodes_to_del = set().union(*all_scc[:-1])

    if nodes_to_del:
        logger.warning(
            f"Removing {len(nodes_to_del)} nodes that form {len(all_scc) - 1} trap components. "
            f"These are groups where you can enter but can't exit (or vice versa). "
            f"Keeping the largest strongly connected component ({len(all_scc[-1])} nodes)."
        )
        graph.remove_nodes_from(nodes_to_del)

    return graph


def estimate_crs_for_bounds(minx, miny, maxx, maxy):
    x_center = np.mean([minx, maxx])
    y_center = np.mean([miny, maxy])
    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(
            west_lon_degree=x_center,
            south_lat_degree=y_center,
            east_lon_degree=x_center,
            north_lat_degree=y_center,
        ),
    )
    crs = CRS.from_epsg(utm_crs_list[0].code)
    logger.debug(f"Estimated CRS for territory {crs}")
    return crs
